/*
 * Copyright (c) 2002, Regents of the University of California. All rights
 * reserved.
 * 
 * =================================================================== The
 * Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if any, must
 * include the following acknowledgment: "This product includes software
 * developed by the Institute for Software Research at University of California,
 * Irvine" Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and "University of
 * California" must not be used to endorse or promote products derived from this
 * software without prior written permission. For written permission, please
 * contact rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called "YANCEES", nor may
 * "YANCEES" appear in their name, without prior written permission of the
 * University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
package edu.uci.isr.yancees;

/**
 * <p>
 * Title: Yancees Notification Server
 * </p>
 * <p>
 * Description: Yet ANother Configurable Extensible Event Service
 * </p>
 * <p>
 * Copyright: Copyright (c) 2003
 * </p>
 * <p>
 * Company: School of Information and Computer Science - University of
 * California, Irvine
 * </p>
 * 
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */

import edu.uci.isr.yancees.core.ProtocolManager;
import edu.uci.isr.yancees.core.ProtocolManagerException;
import edu.uci.isr.yancees.plugin.ProtocolPluginInterface;

/**
 * This class deals with generic messages, that are parserd by installed
 * protocol plug-ins. For example, mobility, authentication and so on.
 * 
 * This class stays behind the YanceesFacade class and deals specifically with
 * protocol manipulation operations.
 *  
 */
public class ProtocolFacade  {

   private ProtocolManager protocolManager = null;

   // single instance of this class
   static ProtocolFacade myInstance;

   // can only be accessed by the YancessAPI or anyone else inside yancess
   // package
   static protected ProtocolFacade getInstance() {
      if (myInstance == null) {
         myInstance = new ProtocolFacade();
      }
      return myInstance;
   }

   // private here prevents direct instantiation of this singleton
   private ProtocolFacade() {
   }

   /**
    * Connects to a protocol object under the name protocolId
    * 
    * @param protocolId
    *           is the name of the protocol plug-in to connect to
    * @return a reference to the plug-in created to handle the protocol, or
    *         throws an exception if an error or no protoocol plug-in found. It
    *         pretty much works as a binding to the protocol object through a
    *         name server such as rmid.
    */
   public synchronized ProtocolPluginInterface connectToSharedProtocol(String protocolId, SubscriberInterface si)
         throws YanceesException {
      
   	ProtocolPluginInterface plugin;

      try {
         plugin = protocolManager.getSharedPluginInstance(protocolId, si);
      } catch (ProtocolManagerException e) {
         e.printStackTrace();
         throw new YanceesException("ProtocolFacade: could not create plugin: "
               + e);
      }

      return plugin;
   }
   
   /**
    * Connects to a protocol object under the name protocolId
    * 
    * @param protocolId
    *           is the name of the protocol plug-in to connect to
    * @return a reference to the plug-in created to handle the protocol, or
    *         throws an exception if an error or no protoocol plug-in found. It
    *         pretty much works as a binding to the protocol object through a
    *         name server such as rmid.
    */
   public synchronized ProtocolPluginInterface connectToNewProtocol(String protocolId, SubscriberInterface si)
         throws YanceesException {
      ProtocolPluginInterface plugin;

      try {
         plugin = protocolManager.getNewUnsharedPluginIntance(protocolId, si);
      } catch (ProtocolManagerException e) {
         e.printStackTrace();
         throw new YanceesException("ProtocolFacade: could not create plugin: "
               + e);
      }

      return plugin;

   }
   
   /**
    * This method informs that the protocol object is not necessary anymore,
    * so it can be garbage collected
    * 
    * @param protocolId
    *           is the name of the protocol plug-in to connect to
    * @param si is the subscriber interface
    * 	
    */
   public void disconnectFromProtocol(String protocolId, SubscriberInterface si)
         throws YanceesException {
      //TODO: implement
   }
   

   /**
    * Used by the Yancees API to initialize this API.
    * 
    * @param pm
    *           is the instance of the protocol manager to be used by this
    *           component
    */
   protected void setProtocolManager(ProtocolManager pm) {
      protocolManager = pm;
   }

}