/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 package edu.uci.isr.yancees;

/**
 * <p>Title: Yancees Notification Server</p>
 * <p>Description: Yet ANother Configurable Extensible Event Service</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: School of Information and Computer Science - University of California, Irvine</p>
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */

import edu.uci.isr.yancees.dispatcher.DispatcherException;
import edu.uci.isr.yancees.dispatcher.EventDispatcherInterface;
import edu.uci.isr.yancees.filter.FilterManagerInterface;

/**
 * This is the server-side publication faade. It hides the yancees core 
 * from the publishers.
 * 
 * It also invokes the appropriate input filters if they are installed
 * 
 * This class stays behind the YanceesServer class
 */
public class PublicationFacade  {

	private EventDispatcherInterface dispatcher = null;
		
	private boolean filtersInstalled = false; // indicate whether the filterManager is installed or not 
	private FilterManagerInterface filterManager;

	// single instance of this class
	static PublicationFacade myInstance;

	// can only be accessed by the YancessAPI or anyone else inside yancess package
	static protected PublicationFacade getInstance() {
		if (myInstance == null) {
			myInstance = new PublicationFacade();
		}
		return myInstance;
	}

	// prevents the direct instantiation of this singleton.
	private PublicationFacade() {

	}

	/**
	 * Installs the inputqueue managed by filterManager in this API
	 * @param filterManager
	 */
	protected void installInputFilters(FilterManagerInterface filterMan) {
		this.filterManager = filterMan;
		filtersInstalled = true;
	}

	/**
	 * Allows the users of this API to publish individual events
	 * @param evt is the event to be published
	 * @throws YanceesException is thrown in case of syntax and connection errors
	 */
	public void publish(EventInterface evt) throws YanceesException {
		
		if (evt != null) {
		
			// All the incoming events are tagged with the timestamp they arrived
			// in the server.
			((GenericMessage) evt).setDateReceivedInServer(new java.util.Date());
	
			if (filtersInstalled) {
				doFilterAndPublish(evt);
			} else {
				publishThroughDispatcher(evt);
			}
		}

	}

	/**
	 * Passes the event trough all the input filtes before publishing it in the
	 * dispatcher
	 * @param evt is the event to be filtered
	 * @throws YanceesException
	 */
	private void doFilterAndPublish(EventInterface evt)
		throws YanceesException {
		EventInterface[] resultList = filterManager.filterEvent(evt);
		publishThroughDispatcher(resultList);
	}

	/**
	 * Publishes an event using the installed event dispatcher. Can be used to bypass
	 * the filters
	 * @param evt is the event to be published
	 * @throws YanceesException
	 */
	protected void publishThroughDispatcher(EventInterface evt)
		throws YanceesException {
		if (dispatcher == null)
			throw new YanceesException("PublisherAPI: Dispatcher needs to be initialized");

		try {
			dispatcher.publish(evt);
		} catch (DispatcherException ex) {
			throw new YanceesException(ex.toString());
		}

	}
	
	
	/**
	  * Publishes an event list using the installed event dispatcher.
	  * This method bypasses the dispatcher.
	  * @param evtList is a list of events to be published
	  * @throws YanceesException in case there are some errors 
	  */
	protected void publishThroughDispatcher(EventInterface[] evtList)
		throws YanceesException {
			if (dispatcher == null)
				throw new YanceesException("PublisherAPI: Dispatcher needs to be initialized");

			try {
				
				for (int i = 0; i < evtList.length; i++) {
					dispatcher.publish(evtList[i]);	
				}
				
			} catch (DispatcherException ex) {
				throw new YanceesException(ex.toString());
			}

		}

	/**
	 * Used by the Yancees faade to initialize this API with the appropriate
	 * default dispatcher.
	 * @param disp
	 */
	protected void setEventDispatcher(EventDispatcherInterface disp) {
		dispatcher = disp;
	}


}