/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 package edu.uci.isr.yancees.server.rmi;

/**
 * <p>Title: Yancees Notification Server</p>
 * <p>Description: Yet ANother Configurable Extensible Event Service</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: School of Information and Computer Science - University of California, Irvine</p>
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */

import java.rmi.Remote;
import java.rmi.RemoteException;

import edu.uci.isr.yancees.EventInterface;
import edu.uci.isr.yancees.MessageInterface;
import edu.uci.isr.yancees.SubscriptionInterface;
import edu.uci.isr.yancees.YanceesException;

/**
 * 
 * @author Roberto Silveira Silva Filho [rsilvafi@ics.uci.edu]
 *
 * This remote interface is the faade for the services in the YANCEES server.
 */
public interface RemoteYanceesInterface extends Remote {

   public static String RMI_LOOKUP_NAME = "Yancees";
 
   /**
    * Connects to an existing protocol instance. If there is no instance available,
    * YANCEES instantiates a new protocol plug-in. Subsequent calls to this method, for
    * that specific protocol, will return the shared plug-in instance.
    * 
    * @param protocolId is the name of the protocol as registered in the configuration file
    * and as available in the protocol factory
    * @param rsi is the subscriber interface that can be used to receive events from the plug-in
    * @return a reference to the plug-in
    * @throws YanceesException is thrown in case of errors in the creation of the protocol plug-in
    * @throws RemoteException is thrown in case of RMI related issues
    */
   public RemoteProtocolPluginInterface connectToSharedProtocol(String protocolId, RemoteSubscriberInterface rsi)
      throws YanceesException, RemoteException;
   
   /**
    * Connects to a new instance of a protocol plug-in. Each time this method is invoked, 
    * a new instance is created.
    * 
    * @param protocolId is the name of the protocol as registered in the configuration file
    * and as available in the protocol factory
    * @param rsi is the subscriber interface that can be used to receive events from the plug-in
    * @return a reference to the plug-in
    * @throws YanceesException is thrown in case of errors in the creation of the protocol plug-in
    * @throws RemoteException is thrown in case of RMI related issues
    */
   public RemoteProtocolPluginInterface connectToNewProtocol(String protocolId, RemoteSubscriberInterface rsi)
   	throws YanceesException, RemoteException;

   /**
    * Terminates unshared protocol instances, or shared protocol instances if no more clients
    * are connected to them.
    * 
    * @param protocolId is the name of the protocol as registered in the configuration file
    * and as available in the protocol factory
    * @param rsi is the subscriber interface that can be used to receive events from the plug-in
    * @throws YanceesException is thrown in case of errors in the creation of the protocol plug-in
    * @throws RemoteException is thrown in case of RMI related issues
    */
   public void disconnectFromProtocol(String protocolId, RemoteSubscriberInterface rsi)
		throws YanceesException, RemoteException;
   
   /**
    * Publishes an opaque event, which can be a XML file, a hashtable representation or
    * another object that implements the eventInterface.
    * @param evt is the event being published
    * @throws YanceesException
    * @throws RemoteException
    */
   public void publish(EventInterface evt) throws YanceesException, RemoteException;

   /**
    * Publishes an array of opaque events, which can be XML files, hashtable representations or
    * other objects that implements the eventInterface.
    * @param evt is the event being published
    * @throws YanceesException
    * @throws RemoteException
    */
	public void publishBuffer(EventInterface[] evt) throws YanceesException, RemoteException;

	/**
	 * Subscribers to YANCEES events. It uses a message that is usually a XMl subscription
	 * @param sub is the subscription
	 * @param rsi is the remote interface to receive the events when the subscription matches
	 * @throws YanceesException
	 * @throws RemoteException
	 */
   public void subscribe(SubscriptionInterface sub, RemoteSubscriberInterface rsi) throws YanceesException, RemoteException;

	/**
	 * unsubscribes or deactivates all subscriptoins registered under the provided subscriber interface
	 * @param rsi is the remote interface used to receive the events when the subscription matches
	 * @throws YanceesException
	 * @throws RemoteException
	 */
   public void unsubscribe(RemoteSubscriberInterface rsi) throws YanceesException, RemoteException;

	/**
	 * unsubscribes or deactivates the subscription registered under the provided subscriber interface
	 * @param rsi is the remote interface used to receive the events when the subscription matches
	 * @param sub is the subscription message
	 * @throws YanceesException
	 * @throws RemoteException
	 */
   public void unsubscribe(RemoteSubscriberInterface rsi, MessageInterface sub) throws YanceesException, RemoteException;

   public void shutdownYancees() throws YanceesException, RemoteException;

   public void suspendYancees() throws YanceesException, RemoteException;

   public void resumeYancees() throws YanceesException, RemoteException;

}