/*
 * Copyright (c) 2008, Regents of the University of California.
 * All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE INSTITUTE FOR SOFTWARE RESEARCH AT THE UNIVERSITY
 * OF CALIFORNIA, IRVINE, OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

package edu.uci.isr.yancees.server.rmi;

import java.rmi.RemoteException;

import edu.uci.isr.yancees.EventInterface;
import edu.uci.isr.yancees.SubscriberInterface;

/**
 * Performs the mapping between the remote interface and the local subscriber
 * interface, working also as a record of the active subscriptions.
 *
 * We have remote and local subscriberInterfaces to allow the implementation
 * of different protocols, for example HTTP or SOAP, that may heve different
 * ways of representing a remote SubscriberInterface, so we need someone to
 * be a mediator here for the sake of future extensions
 *
 * Behind this faade, yancees interacts only with SubscriberInterfaces
 * whereas, to the outside world, RemoteSubscriberInterfaces are used. Hence,
 * a translation between both is necessary here.
 */
public class SubscriberMediator implements SubscriberInterface {

	// Indicates whether to use or not the notification buffer
	private  boolean USE_NOTIFICATION_BUFFER = edu.uci.isr.yancees.YanceesProperties.getInstance().USE_PUBLICATION_BUFFER;
	
	private RemoteSubscriberInterface _remoteInterface;
	private int _referenceCounter;
	private NotificationBuffer _notificationBuffer;
	private RemoteYanceesImplementation _yancees;
	
	private boolean print = edu.uci.isr.yancees.YanceesProperties.getInstance().PRINT_DEBUG;

	/**
	 * constructor
	 *
	 * @param rsi
	 *           is the remote interface, that came from the publish or
	 *           subscribe commands, protocols as well.
	 * @param callback
	 *           is
	 */
	public SubscriberMediator(RemoteSubscriberInterface rsi,
			RemoteYanceesImplementation ryi) {

		_yancees = ryi;
		_remoteInterface = rsi;
		_referenceCounter = 1; // one

		if (USE_NOTIFICATION_BUFFER) {
			_notificationBuffer = new NotificationBuffer(rsi, this);
			_notificationBuffer.start();
		}

	}
	
	/**
	 * Used by the NotificationBuffer to obtain a reference to Yancees remote API.
	 * @return the server implementation that created this mediator.
	 */
	public RemoteYanceesImplementation getRemoteYanceesReference() {
		return _yancees;
	}

	/**
	 * Some applications need the actual remote interface used in the 
	 * client side. For example, this may be important to uniquely identify
	 * the clients when using protocol plug-ins.
	 * @return the actual remote interface that this local SubscriberInterface represents
	 */
	public RemoteSubscriberInterface getRemoteSubscriberReference() {
		return _remoteInterface;
	}
	/**
	 * Receives notifications from the local components of Yancees and
	 * forwards them to the remote subscribers
	 */
	public void notify(EventInterface evt) {

		if (print) 
			System.out.println("Adding event to buffer...");
		if (USE_NOTIFICATION_BUFFER) {
			_notificationBuffer.addObjectToBuffer(evt);
		} else {
			try {
				_remoteInterface.notify(evt);
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}
		
		/*
		 * try { remoteInterface.notify(evt); } catch (RemoteException ex) { if
		 * (print) { System.out.println( "SubscriberMediator: error connecting
		 * to remote interface"); System.out.println(ex); }
		 */

		/**
		 * This piece of code is here to perform the disposal of Remote
		 * Interfaces that are disconnected, for example, due to a network
		 * error or abrupt client disconnection. Once it happpens, it is
		 * necessary to terminate all the subscriptions performed by that
		 * subscriber as well as unregister this mediator.
		 */
		/*
		 * try { if (print) System.out.println( "SubscriberMediator:
		 * unsubscribing remote interface...");
		 * subscriberAPI.unsubscribe(this); mediatorsTable.remove(this); if
		 * (print) System.out.println( "SubscriberMediator: remote interface
		 * unsubscribed."); } catch (YanceesException e) { System.out.println(
		 * "SubscriberMediator: error when unregistering the remote interface
		 * as a garbage collection strategy"); System.out.println(e); }
		 */
		// }
	}

	/**
	 * Receives notifications from the local components of Yancees and
	 * forwards them to the remote subscribers
	 */
	public void notify(EventInterface[] evtList) {
		
		if (USE_NOTIFICATION_BUFFER) {
			_notificationBuffer.addObjectToBuffer(evtList);
		} else {
			try {
				_remoteInterface.notify(evtList);
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		/*
		 * try { remoteInterface.notify(evtList); } catch (RemoteException ex) {
		 * if (print) { System.out.println( "SubscriberMediator: error
		 * connecting to remote interface"); System.out.println(ex); }
		 */

		/*
		 * try { System.out.println( "SubscriberMediator: unsubscribing remote
		 * interface..."); subscriberAPI.unsubscribe(this);
		 * mediatorsTable.remove(this); System.out.println(
		 * "SubscriberMediator: remote interface unsubscribed."); } catch
		 * (YanceesException e) { System.out.println( "SubscriberMediator:
		 * error when unregistering the remote interface as a garbage
		 * collection strategy"); System.out.println(e); }
		 */

		// }
	}

	public int incReferenceCounter() {
		_referenceCounter++;
		return _referenceCounter;
	}

	public int decReferenceCounter() {
		_referenceCounter--;
		return _referenceCounter;
	}

}